from lolpop.component.resource_version_control.base_resource_version_control import BaseResourceVersionControl
from lolpop.component.metadata_tracker.continual_metadata_tracker import ContinualMetadataTracker
from lolpop.utils import common_utils as utils
from lolpop.utils import continual_utils as cutils
import os
import joblib
import pandas as pd

@utils.decorate_all_methods([utils.error_handler,utils.log_execution()])
class ContinualVersionControl(BaseResourceVersionControl): 
    __REQUIRED_CONF__ = {
        "config" : ["local_dir"]
    }
    def __init__(self, description=None, run_id=None, components = {}, *args, **kwargs): 
        #set normal config
        super().__init__(components=components, *args, **kwargs)
        
        # if we are using continual for metadata tracking then we won't have to set up connection to continual
        # if not, then we do. If would be weird to have to do this, but just in case. 
        if isinstance(components.get("metadata_tracker"), ContinualMetadataTracker): 
            self.client = self.metadata_tracker.client
            self.run = self.metadata_tracker.run
        else: 
            secrets = utils.load_config(["CONTINUAL_APIKEY", "CONTINUAL_ENDPOINT", "CONTINUAL_PROJECT", "CONTINUAL_ENVIRONMENT"], self.config)
            self.client = cutils.get_client(secrets)
            self.run = cutils.get_run(self.client, description=description, run_id=run_id)

    def version_data(self, dataset_version, data, key = "data_csv", **kwargs): 
        id = self.metadata_tracker.get_resource_id(dataset_version)

        #dump dataframe to local
        local_path = "%s/%s.csv" %(self.config.get("local_dir"),id)
        data.to_csv(local_path, index=False)

        artifact = self.metadata_tracker.log_artifact(dataset_version, id = key, path = local_path, mime_type="csv", external=False)
        
        return {"uri" : artifact.url}

    def get_data(self, dataset_version, vc_info=None, key="data_csv", **kwargs):
        artifact = dataset_version.artifacts.get(id=key)
        file_path = "%s/%s/%s" %(self._get_config("local_dir"),dataset_version.id, artifact.id)
        os.makedirs(file_path, exist_ok=True)
        _, download_path = artifact.download(dest_dir=file_path)
        df = pd.read_csv(download_path)
        return df

    def version_model(self, experiment, model, algo, **kwargs): 
        id = self.metadata_tracker.get_resource_id(experiment)

        model_dir = "%s/models/%s" %(self._get_config("local_dir"), algo)    
        os.makedirs(model_dir, exist_ok=True)
        model_path = "%s/%s" %(model_dir, id)
        joblib.dump(model, model_path)
        artifact = self.metadata_tracker.log_artifact(experiment, id = "model_artifact", path=model_path, external=False)

        return {"uri" : artifact.url}
    
    def get_model(self, experiment, key="model_artifact", **kwargs): 
        artifact = self.metadata_tracker.get_artifact(experiment,id=key)
        file_path = "%s/%s" %(self._get_config("local_dir"),experiment.id)
        os.makedirs(file_path, exist_ok=True)
        _, download_path = artifact.download(dest_dir=file_path)
        model = joblib.load(download_path)
        
        return model 
 


from lolpop.pipeline.process.base_process import BaseProcess
from lolpop.utils import common_utils as utils
from lolpop.utils import metaflow_utils as meta_utils
from metaflow import FlowSpec, step
from pathlib import Path

METAFLOW_CLASS = "MetaflowOfflineProcessSpec"
PLUGIN_PATHS = "plugin_paths.txt"


@utils.decorate_all_methods([utils.error_handler, utils.log_execution()])
class MetaflowOfflineProcess(BaseProcess):
    __REQUIRED_CONF__ = {
        "components": ["data_transformer", "metadata_tracker", "resource_version_control", "data_profiler", "data_checker"],
        "config": []
    }

    def run(self, source_data_name, source_data, **kwargs):
        #get flow class object from this file
        mod_cl = meta_utils.get_flow_class(__file__, METAFLOW_CLASS)

        flow = meta_utils.load_flow(mod_cl, self, PLUGIN_PATHS, source_data=source_data, source_data_name=source_data_name)
        self.log("Loaded metaflow flow %s" % METAFLOW_CLASS)

        meta_utils.run_flow(flow, "run", __file__, PLUGIN_PATHS)
        self.log("Metaflow pipeline %s finished." % METAFLOW_CLASS)

    def get_artifacts(self, artifact_keys):
        #get latest run of this pipeline
        run = meta_utils.get_latest_run(METAFLOW_CLASS)

        #get requested artifacts
        artifacts = meta_utils.get_run_artifacts(run, artifact_keys, METAFLOW_CLASS)

        return artifacts
class MetaflowOfflineProcessSpec(FlowSpec):

    def __init__(self, lolpop=None, source_data=None, source_data_name=None, use_cli=False, **kwargs):
        #you need to set local attributes before calling super
        #only the first time we create the class will these parameters be provided.
        #The rest of the calls are metaflow internal calls and we do not want to reset them
        if lolpop is not None:
            self.lolpop = lolpop
        if source_data is not None:
            self.source_data = source_data
        if source_data_name is not None: 
            self.source_data_name = source_data_name

        #load plugins
        meta_utils.load_plugins(PLUGIN_PATHS)

        #call init -- this kicks off the main workflow
        super().__init__(use_cli=use_cli)

    @step
    def start(self):
        self.next(self.transform_data)

    @step
    def transform_data(self):
        #transform data
        self.data = self.lolpop.data_transformer.transform(
            self.source_data)

        self.next(self.track_data)

    @step
    def track_data(self):
        #create dataset version
        dataset_version = self.lolpop.metadata_tracker.create_resource(
            self.source_data_name, type="dataset_version")
        self.dataset_version = dataset_version
        self.lolpop.datasets_used.append(dataset_version)

        #version data
        vc_info = self.lolpop.resource_version_control.version_data(
            dataset_version, self.data)

        #register version control metadata w/ metadata tracker
        self.lolpop.metadata_tracker.register_vc_resource(
            dataset_version, vc_info, file_type="csv")

        self.next(self.profile_data)

    @step
    def profile_data(self):
        #profile data
        data_profile, file_path = self.lolpop.data_profiler.profile_data(
            self.data)

        #log profile to metadata tracker
        self.lolpop.metadata_tracker.log_data_profile(
            self.dataset_version,
            file_path=file_path,
            profile=data_profile,
            profiler_class=self.lolpop.data_profiler.name
        )
        self.next(self.check_data)

    @step
    def check_data(self):
        #run data checks
        data_report, file_path, checks_status = self.lolpop.data_checker.check_data(
            self.data)

        #log data report to metadata tracker
        self.lolpop.metadata_tracker.log_checks(
            self.dataset_version,
            file_path=file_path,
            report=data_report,
            checker_class=self.lolpop.data_checker.name,
            type="data"
        )

        if checks_status == "ERROR" or checks_status == "WARN":
            url = self.lolpop.metadata_tracker.url
            self.lolpop.notify(
                "Issues found with data checks. Visit %s for more information." % url, checks_status)
        self.next(self.compare_data)

    @step
    def compare_data(self):
        #get previous dataset version & dataframe
        prev_dataset_version = self.lolpop.metadata_tracker.get_prev_resource_version(
            self.dataset_version)

        if prev_dataset_version is not None:
            vc_info = self.lolpop.metadata_tracker.get_vc_info(
                prev_dataset_version)
            prev_data = self.lolpop.resource_version_control.get_data(
                prev_dataset_version, vc_info)

            #compare current dataset version with previous dataset version
            comparison_report, file_path = self.lolpop.data_profiler.compare_data(
                self.data, prev_data)

            self.lolpop.metadata_tracker.log_data_comparison(
                self.dataset_version,
                file_path=file_path,
                report=comparison_report,
                profiler_class=self.lolpop.data_profiler.name
            )
        else:
            self.lolpop.log("No previous dataset version found for dataset: %s" % (
                self.lolpop.metadata_tracker.get_resource_id(self.dataset_version)))

        self.next(self.end)

    @step
    def end(self):
        pass


if __name__ == '__main__':
    #need to use CLI here because metaflow calls back into this function to
    #execute each individual step
    MetaflowOfflineProcessSpec(use_cli=True)



from lolpop.component.hyperparameter_tuner.base_hyperparameter_tuner import BaseHyperparameterTuner
from lolpop.utils import common_utils as utils
import optuna
from datetime import datetime
import os

@utils.decorate_all_methods([utils.error_handler,utils.log_execution()])
class OptunaHyperparameterTuner(BaseHyperparameterTuner): 

    def run_experiment(self, data, model_version, n_trials=100, *args, **kwargs): 
        #set up params
        training_params = self._get_config("training_params")
        trainer_configs = self._get_config("trainer_configs", {})
        param_type = self._get_config("param_type", "fixed")
        metrics = self._get_config("metrics")
        perf_metric = self._get_config("perf_metric")
        # default to 1 hr timeout
        timeout = self._get_config("optuna_timeout", 3600)
        n_jobs = self._get_config("num_jobs", 1)

        #understand if we want to minimize or maximum objective for optuna
        reverse = utils.get_metric_direction(perf_metric)
        direction = "maximize"
        if not reverse: 
            direction = "minimize"
 
        exp_list = {} 
        model_list = {}
        for algo in training_params: 
            #need to create study w/ sampler so that the results are reproducible later on
            sampler = optuna.samplers.TPESampler(seed=42) 
            study_name = self.metadata_tracker.get_resource_id(model_version)
            study = optuna.create_study(direction=direction, sampler=sampler, study_name=study_name)
            if param_type == "fixed":
                grid = self._build_training_grid(training_params[algo])
                #only run n_trials = len(grid) to only run enqued params
                n_trials = len(grid)
                self.log("Found %s parameter combinations. Proceeding to enqueue all combinations..." %n_trials)
                #enqueue all fixed params
                for params in grid: 
                    study.enqueue_trial(params)   
            #run study
            study.optimize(lambda trial: self._optuna_objective(trial, param_type, data, model_version, algo, training_params[algo], trainer_configs.get(algo,{}), metrics, perf_metric, exp_list, model_list), n_trials=n_trials, timeout=timeout, n_jobs=n_jobs)
            #log study
            #study.set_user_attr("perf_metric_name", perf_metric)
            #study.set_user_attr("perf_metric_val", study.best_value)
            self._log_study(study, model_version, algo) #returns best exp id 
            #get best experiment in study
            #best_experiment = study.best_trial.user_attrs.get("experiment_id")
            #exp_model = study.best_trial.user_attrs.get("model")
            #exp_list[best_experiment] = study.best_value
            #model_list[best_experiment] = exp_model 

        #save data splits
        for k,v in data.items(): 
            vc_info = self.resource_version_control.version_data(model_version, v, key=k)
            self.metadata_tracker.register_vc_resource(model_version, vc_info, key=k, file_type="csv")

        #now, we determine overall best experiment and save into model_version
        winning_exp_id = self._get_winning_experiment(exp_list, perf_metric, reverse=reverse)
        winning_exp = self.metadata_tracker.get_resource(winning_exp_id, parent=model_version, type="experiment")
        winning_exp_model_trainer = self.metadata_tracker.get_metadata(winning_exp, id="model_trainer")
        best_model = model_list.get(winning_exp_id)

        #log important stuff to model version
        self.metrics_tracker.copy_metrics(winning_exp, model_version)
        self.metadata_tracker.log_metadata(model_version, id="winning_experiment_id", data=winning_exp_id)
        self.metadata_tracker.log_metadata(model_version, id="winning_experiment_model_trainer", data=winning_exp_model_trainer)
      
        return best_model

    def _optuna_objective(self, trial, param_type, data, model_version, algo, params, trainer_config, metrics, perf_metric, experiment_list, model_list): 
        if param_type == "dynamic": 
            model_params = self._get_dynamic_params(trial, params)
        else: #param_type=="fixed"
            model_params = self._get_fixed_params(trial, params)
        #build model
        model, experiment = self.build_model(data, model_version, algo, model_params, trainer_config)
        #make predictions
        predictions = model.predict(data)
        #calculate metrics
        metrics_val = model.calculate_metrics(data, predictions, metrics)
        perf_metric_value = metrics_val["valid"][perf_metric]

        #need to save exp model now because optuna doesn't provide access to models objects later on. 
        #use model_params instaed of trial.params as optuna doesn't track fixed parameters, but the user may want to
        self._log_trial(trial, model_params, model, experiment, algo)
        #log metrics
        self.metrics_tracker.log_metrics(experiment, metrics_val, perf_metric)

        #save experiment id and model so we can refer back to it later 
        #note, for large models this may cause memory issues!
        experiment_id = self.metadata_tracker.get_resource_id(experiment)
        experiment_list[experiment_id] = perf_metric_value
        model_list[experiment_id] = model
        
        #clean up resource
        self.metadata_tracker.clean_resource(experiment, type="experiment")

        #objective function needs to return perf value for so optuna can do comparison
        return perf_metric_value

    # parses each value of a dynamic config to the optuna type
    def _parse_dynamic_logic(self, trial, name, p):
        param_type = p.get("type")

        if param_type == "fixed": 
            return p.get("value")
        elif param_type == "int": 
            low_high = p.get("range")
            return trial.suggest_int(name, low_high[0], low_high[1])
        elif param_type == "float":
            low_high = p.get("range")
            return trial.suggest_float(name, low_high[0], low_high[1])
        elif param_type == "categorical":
            choices = p.get("choices")
            return trial.suggest_categorical(name, choices)

    def _get_dynamic_params(self, trial, params):
        params_out = {}
        for p in params: 
            params_out[p] = self._parse_dynamic_logic(trial, p, params[p])
        return params_out

    def _get_fixed_params(self, trial, params): 
        params_out = {}
        for p in params: #since the params passed for a fixed set are static, we can represent them as a categorical
            params_out[p] = trial.suggest_categorical(p, params[p])
        return params_out

    def _log_trial(self, trial, model_params, model, experiment, algo):
        #log params + trainer
        self.metadata_tracker.log_metadata(experiment, id="optuna_trial_number", data=trial.number)
        self.metadata_tracker.update_resource(experiment, {"start_time": trial.datetime_start, "end_time":datetime.utcnow()})
        
        #save model
        self.save_model(model, experiment, model_params, algo)

        return experiment

    #saves interesting parts of the study
    def _log_study(self, study, model_version, algo): 
        if optuna.visualization.is_available(): 
            plot = optuna.visualization.plot_edf(study)
            self._save_plot("optuna_plot_edf.html",plot, model_version,algo)

            #only useful if using intermediate values and pruning
            #plot = optuna.visualization.plot_intermediate_values(study)
            #self._save_plot("optuna_plot_intermediate_values.html",plot, model_version,algo)

            plot = optuna.visualization.plot_optimization_history(study)
            self._save_plot("optuna_plot_optimization_history.html",
                            plot, model_version, algo)

            plot = optuna.visualization.plot_parallel_coordinate(study)
            self._save_plot("optuna_plot_parallel_coordinate.html",
                            plot, model_version, algo)

            plot = optuna.visualization.plot_param_importances(study)
            self._save_plot("optuna_plot_param_importances.html",
                            plot, model_version, algo)

            #only used for multi-objective studies
            #optuna.visualization.plot_pareto_front(study)
            #self._save_plot("optuna_plot_pareto_front.html",plot, model_version,algo)

            plot = optuna.visualization.plot_slice(study)
            self._save_plot("optuna_plot_slice.html",
                            plot, model_version, algo)

            plot = optuna.visualization.plot_contour(study)
            self._save_plot("optuna_plot_contour.html",
                            plot, model_version, algo)
    
    #saves plot 
    def _save_plot(self, name, plot, model_version, algo): 
        name_arr = name.split(".")
        plot_type = name_arr[-1]
        plot_name = "_".join(name_arr)

        local_path = "%s/%s/%s" %(
            self._get_config("local_dir"), 
            self.metadata_tracker.get_resource_id(model_version),
            algo
            )
        os.makedirs(local_path, exist_ok=True)
        local_file = "%s/%s" %(local_path, name)
        plot.write_html(local_file)
        artifact = self.metadata_tracker.log_artifact(model_version, id = plot_name, path = local_file, mime_type = plot_type, external = False)

        return artifact


import typer 
from typer.core import TyperGroup


class NaturalOrderGroup(TyperGroup):
    def list_commands(self, ctx):
        return self.commands.keys()
    
app = typer.Typer(cls = NaturalOrderGroup, help="Package a lolpop workflow.")


@app.callback(invoke_without_command=True)
def default(ctx: typer.Context):
    if ctx.invoked_subcommand is not None:
        return
    else:
        typer.echo(ctx.command.get_help(ctx))



from lolpop.component.resource_version_control.base_resource_version_control import BaseResourceVersionControl
from lolpop.utils import common_utils as utils
import os 
import joblib
import pandas as pd

# an assumption we make using this component is that dvc is already set up. 
# that is -- `dvc init --subdir` has been run on a subdir of the main git repo
# this is specified by the DVC_DIR variable. And a remote has been set up with 
# the id given in DVC_REMOTE 
@utils.decorate_all_methods([utils.error_handler,utils.log_execution()])
class dvcVersionControl(BaseResourceVersionControl): 
    __REQUIRED_CONF__ = {
        "config" : ["DVC_DIR", "DVC_REMOTE"]
    }

    __DEFAULT_CONF__ = {
        "config": {"dvc_dir": "dvc/", "dvc_remote": "local"}
    }

    def __init__(self, *args, **kwargs): 
        #set normal config
        super().__init__(*args, **kwargs)
        
        secrets = utils.load_config(["DVC_DIR", "DVC_REMOTE"], 
                                    utils.copy_config_into(self.config,self.__DEFAULT_CONF__.get("config",{})))

        self.dvc_dir = secrets.get("DVC_DIR")
        if self.dvc_dir[-1] != "/":
            self.dvc_dir = self.dvc_dir + "/"
        self.dvc_remote = secrets.get("DVC_REMOTE")

        if self.dvc_remote == "local": 
            output, _ = utils.execute_cmd(["dvc",  "remote", "list"])
            if "local" not in output: 
                local_dir = "%s/dvc" %(self._get_config("local_dir"))
                self.log("Creating dvc remote 'local' for local directory %s" %local_dir)
                _ = utils.execute_cmd(["dvc", "remote", "add", "-d", "local",  local_dir])

        self.git_url = utils.execute_cmd(["git", "config", "--get", "remote.origin.url"], self)[0].strip()

    def version_data(self, dataset_version, data, key = None, **kwargs): 
        id = self.metadata_tracker.get_resource_id(dataset_version)

        #set up paths
        dvc_path = self.dvc_dir
        if key: 
            dvc_file = "%s_%s.csv" %(id, key)
        else: 
            dvc_file = "%s.csv" %(id)
        file_path = dvc_path + dvc_file
        
        #dump dataframe to dvc path
        data.to_csv(file_path, index=False)
        
        #commit file to dvc and .dvc file to git, and  then dvc push
        _,_ = utils.execute_cmd(["dvc", "add", file_path], self)
        _,_ = utils.execute_cmd(["dvc",  "commit", file_path], self)
        #git repo needs to have .dvc directory or else dvc doesn't know where to download from when using `dvc get`
        hexsha = utils.git_commit_file(".dvc/", msg="Committing .dvc directory", push=False, logger=self)
        hexsha = utils.git_commit_file("%s.dvc" %file_path, msg="Committing dvc file", logger=self)
        _,_ = utils.execute_cmd(["dvc", "push", "--remote", self.dvc_remote], self)
        #I don't think the URI is really needed. We should always access via dvc commands 
        URI, _ = utils.execute_cmd(["dvc", "get", os.getcwd(), file_path, "--show-url"], self)
        
        return {"uri" : URI, "hexsha": hexsha}

    def get_data(self, dataset_version, vc_info, key = None, **kwargs):
        if vc_info is not None: 
            hexsha = vc_info.get("hexsha")
        id = self.metadata_tracker.get_resource_id(dataset_version)
        if key:
            dvc_file = "%s_%s.csv" % (id, key)
        else:
            dvc_file = "%s.csv" % (id)

        file_path = self.dvc_dir + dvc_file 
        #if path exists, remove it before downloading. 
        if os.path.exists(dvc_file):
            os.remove(dvc_file)
        _, _ = utils.execute_cmd(["dvc", "get", self.git_url, file_path, "--rev", hexsha, "-o", dvc_file], self)
        df = pd.read_csv(dvc_file)
        
        return df 

    def version_model(self, experiment, model, algo, key=None, **kwargs): 
        id = self.metadata_tracker.get_resource_id(experiment)
        model_version_id = self.metadata_tracker.get_parent_id(experiment, type="experiment")
        if key:
            id = "%s_%s" % (id, key)

        # set up directories and dump model
        dvc_path = self.dvc_dir
        model_dir = dvc_path + "models/%s/%s" %(model_version_id, algo)    
        os.makedirs(model_dir, exist_ok=True)
        model_file = "models/%s/%s/%s" %(model_version_id, algo, id)
        model_path = dvc_path + model_file  
        joblib.dump(model, model_path)

        #now commit to dvc
        _,_ = utils.execute_cmd(["dvc", "add", model_path], self)
        _,_ = utils.execute_cmd(["dvc", "commit", model_path], self)
        hexsha = utils.git_commit_file(".dvc/", msg="Committing .dvc directory", push=False, logger=self)
        hexsha = utils.git_commit_file("%s.dvc" %model_path,msg="Committing dvc file", logger=self)
        _,_ = utils.execute_cmd(["dvc", "push", "--remote", self.dvc_remote], self)
        URI, _ = utils.execute_cmd(["dvc", "get", os.getcwd(), model_path, "--show-url"], self)

        return {"uri" : URI, "hexsha": hexsha}
  
    def get_model(self, experiment, key = None, **kwargs): 
        id = self.metadata_tracker.get_resource_id(experiment)
        if key:
            id = "%s_%s" % (id, key)
        model_version_id = self.metadata_tracker.get_parent_id(experiment, type="experiment")
        algo = self.metadata_tracker.get_metadata(experiment,id="model_trainer")
        hexsha = self.metadata_tracker.get_vc_info(experiment, key="hexsha").get("hexsha")
        dvc_file = "models/%s/%s/%s" %(model_version_id,algo,id)
        file_path = self.dvc_dir + dvc_file
        #if path exists, remove it before downloading.
        if os.path.exists(dvc_file):
            os.remove(dvc_file)
        _, _ = utils.execute_cmd(["dvc", "get", self.git_url, file_path,  "--rev", hexsha, "-o", dvc_file], self)
        model = joblib.load(dvc_file)
        return model

    


import os 
import json
from continual import Client 

# Get continual client
def get_client(config): 
    client = Client(
        api_key=config.get("CONTINUAL_APIKEY"), 
        endpoint=config.get("CONTINUAL_ENDPOINT"),
        project=config.get("CONTINUAL_PROJECT"),
        environment=config.get("CONTINUAL_ENVIRONMENT"),
    )
    return client


# Get or return continual run
def get_run(client, description=None, run_id=None): 
    try: 
        run = client.runs.create(description=description, id=run_id)
    except: 
        pass 

    #create doesn't actually error, it just returns None if the resource exists. 
    if run is None: 
        run = client.runs.get(run_id)
    return run 


# Get continual dataset
## Returns dataset,dataset_version as the latter is what you'll likely really want to work with
def get_dataset(run, id, create_dataset_version=False): 
    dataset = run.datasets.create(id)
    dataset_version = None
    if create_dataset_version: 
        dataset_version = dataset.dataset_versions.create()
    return (dataset, dataset_version)

def get_dataset_version(run, id): 
    dataset_version = run.dataset_versions.get(id)


def convert_to_metrics_dicts(metrics): 
    metrics_dicts = []
    value_dicts = []
    for k1 in metrics.keys(): #train/valid/test
        for k2 in metrics[k1].keys(): #metric key
            metrics_dicts.append(
                {
                    "id": k2,
                    "value": metrics[k1][k2],
                    "group": k1,
                }
            )

    return metrics_dicts



from lolpop.component.model_trainer.base_model_trainer import BaseModelTrainer
from lolpop.utils import common_utils as utils
from prophet import Prophet
from matplotlib import pyplot as plt 

import pandas as pd 
import numpy as np

@utils.decorate_all_methods([utils.error_handler, utils.log_execution()])
class ProphetModelTrainer(BaseModelTrainer):

    mlflow_module = "prophet"

    #should set self.model in init
    def __init__(self, problem_type=None, params={}, *args, **kwargs):
        #set normal config
        super().__init__(problem_type=problem_type, params=params, *args, **kwargs)
        self.model = Prophet(**params)

    def fit(self, data, *args, **kwargs):
        self.log("Starting model training with parameters: %s" %
                 str(self.params))

        df = data.get("X_train")
        df["y"] = data.get("y_train")
    
        df_train = self._process_data(df)

        country_holidays = self._get_config("country_holidays")
        if country_holidays:
            self.model.add_country_holidays(country_name=country_holidays)

        #add regressors
        regressors = self._get_config("regressor_cols")
        if regressors is not None: 
            regressor_list = regressors.split(",")
            for regressor in regressor_list:
                self.model.add_regressor(regressor)

        self.model.fit(df_train)

        return self.model

    def predict(self, data, *args, **kwargs):
        predictions={}
        target_min = self._get_config("target_min")
        target_max = self._get_config("target_max")


        df_train = self._process_data(data["X_train"])
        predictions["train_confidence"] = self._apply_prediction_bounds(
            self.model.predict(df_train), target_min, target_max)
        predictions["train"] = predictions["train_confidence"]["yhat"]
        

        if data.get("X_valid") is not None:
            df_valid= self._process_data(data["X_valid"])
            predictions["valid_confidence"] = self._apply_prediction_bounds(
                self.model.predict(df_valid), target_min, target_max)
            predictions["valid"] = predictions["valid_confidence"]["yhat"]

        if data.get("X_test") is not None:
            df_test = self._process_data(data["X_test"])
            predictions["test_confidence"] = self._apply_prediction_bounds(
                self.model.predict(df_test), target_min, target_max)
            predictions["test"] = predictions["test_confidence"]["yhat"]

        return predictions

    #prophet doesn't seem to do a good job of respecting defined floor/cap on historic data
    #this can cause errors with metrics when numbers are negative, etc 
    def _apply_prediction_bounds(self, data, target_min=None, target_max=None): 
        data_out = data.copy()
        if target_min is not None: 
            data_out["yhat"] = data["yhat"].apply(lambda x: target_min if x < target_min else x)
        if target_max is not None: 
            data_out["yhat"] = data["yhat"].apply(lambda x: target_max if x > target_max else x) 
        return data_out 


    #def calculate_metrics(self, data, predictions, metrics, **kwargs):
    #    for key in predictions.keys(): 
    #        predictions[key]=predictions[key]["yhat"].to_numpy()
    #    return super().calculate_metrics(data, predictions, metrics, **kwargs) 

    def predict_df(self, df):
        data = self._process_data(df)
        return self.model.predict(data)["yhat"]

    def _process_data(self, data): 
        #prophet expects labels are in a column 'y' and timestamp is 'ds'
        df_train = data.rename(columns={
            self._get_config("time_index"): "ds",
            self._get_config("model_target"): "y",
        }, errors="ignore")

        #set min/max values
        target_min = self._get_config("target_min", None)
        target_max = self._get_config("target_max", None)
        if target_min: 
            df_train["floor"] = target_min
        if target_max: 
            df_train["cap"] = target_max

        data_out = df_train.sort_values("ds").reset_index(drop=True)

        return data_out
    
    def get_artifacts(self, id): 
        artifacts_out  = {}

        # generate component plot
        historic_predictions = self.model.predict(self.model.history)
        self.model.plot_components(historic_predictions)
        local_dir = self._get_config("local_dir")
        file_path = "%s/%s_%s" % (local_dir, id, "component_plot.png")
        plt.savefig(file_path)
        artifacts_out["component_plot"] = file_path 
     
        #plot forecast
        future = self.model.make_future_dataframe(
            periods=self._get_config("forecast_period"), 
            freq=self._get_config("forecast_frequency"),
            include_history=True
            )
        predictions = self.model.predict(future)
        self.model.plot(predictions)
        file_path = "%s/%s_%s" %(local_dir, id, "forecast_plot.png")
        plt.savefig(file_path)
        artifacts_out["forecast_plot"] = file_path

        return artifacts_out



from lolpop.pipeline.base_pipeline import BasePipeline

class BaseWrapper(BasePipeline):
    def __init__(self, pipeline_type="wrapper", *args, **kwargs):
        super().__init__(pipeline_type=pipeline_type, *args, **kwargs)

    def run(self, *args, **kwargs):
        pass


from lolpop.component.data_checker.base_data_checker import BaseDataChecker
from lolpop.utils import common_utils as utils
from evidently.test_suite import TestSuite
from evidently.test_preset import *

@utils.decorate_all_methods([utils.error_handler,utils.log_execution()])
class EvidentlyAIDataChecker(BaseDataChecker): 
    
    __REQUIRED_CONF__ = {
        "config" : ["local_dir"]
    }

    def check_data(self, data, *args, **kwargs): 
        """Generates a data check report using EvidentlyAI.

        Args:
            data (pd.DataFrame): A dataframe of the data to check

        Returns:
            data_report (object): Python object of the data report.
            file_path (string):  Path to the exported report. 
            checks_status (string): Status of the checks ("PASS"/"WARN"/"ERROR", etc.)
        """
        data_report = TestSuite(tests=[DataQualityTestPreset()])
        data_report.run(current_data=data, reference_data=None)
        file_path = "%s/EVIDENTLY_DATA_REPORT.HTML" %self._get_config("local_dir")
        data_report.save_html(file_path)

        summary = data_report.as_dict()["summary"]
        
        num_checks_passed = summary["success_tests"]
        num_checks_failed = summary["failed_tests"]
        num_checks_not_ran = summary["total_tests"] - (num_checks_failed + num_checks_passed)
        self.log("%s had %s passed checks." % (self.name, num_checks_passed))
        self.log("%s had %s failed checks." % (self.name, num_checks_failed))
        self.log("%s had %s checks not run." % (self.name, num_checks_not_ran))

        checks_status = "PASS"
        if num_checks_failed > 0: 
            checks_status = "ERROR"
        elif num_checks_not_ran > 0: 
            checks_status = "WARN"

        return data_report, file_path, checks_status



from lolpop.component.base_component import BaseComponent


class BaseDataSynthesizer(BaseComponent):

    def load_data(self, source_file, *args, **kwargs):
        pass

    def model_data(self, data, *args, **kwargs):
        pass

    def sample_data(self, model, num_rows, *args, **kwargs):
        pass

    def evaluate_data(self, real_data, synthetic_data, *args, **kwargs):
        pass
